from abc import ABC
from contextlib import suppress, contextmanager
from collections.abc import Generator
from typing import TYPE_CHECKING, Optional

from pydantic import ValidationError
import httpx

if TYPE_CHECKING:
    from .api_methods import ErrorResponse


class UnisenderGoError(Exception, ABC):
    pass


@UnisenderGoError.register
class ClientSetupError(Exception, ABC):
    pass


@ClientSetupError.register
class SyncClientSetupError(Exception):
    def __init__(self) -> None:
        super().__init__('Requires SyncClient to be setup before call.')


@ClientSetupError.register
class AsyncClientSetupError(Exception):
    def __init__(self) -> None:
        super().__init__('Requires AsyncClient to be setup before call.')


@UnisenderGoError.register
class HTTPStatusError(httpx._exceptions.HTTPStatusError):
    # Common httpx.HTTPStatusError fields are filled by __init__ method:
    request: httpx.Request
    response: httpx.Response

    # Unisender Go specific extra data:
    response_payload: Optional['ErrorResponse']

    def __init__(
        self,
        *,
        request: httpx.Request,
        response: httpx.Response,
    ) -> None:
        from .api_methods import ErrorResponse

        self.response_payload = None
        with suppress(ValidationError):
            self.response_payload = ErrorResponse.parse_raw(response.content)

        message_template_lines = [
            "Error code={payload.code!r} message={payload.message!r};" if self.response_payload else None,
            "Error occured on {request.method} {request.url}",
            "Response HTTP status is {response.status_code} {response.reason_phrase}",
            "Response payload: {response.text}",
            'For more information check: https://godocs.unisender.ru/web-api-ref#api-errors',
        ]
        message_template = '\n'.join(line for line in message_template_lines if line)

        message = message_template.format(
            payload=self.response_payload,
            request=request,
            response=response,
        )

        super().__init__(message, request=request, response=response)


@UnisenderGoError.register
class ResponseFormatError(Exception):
    request: httpx.Request
    response: httpx.Response

    def __init__(
        self,
        *,
        request: httpx.Request,
        response: httpx.Response,
    ):
        self.request = request
        self.response = response
        exception_message = (
            f"Error occured on {request.method} {request.url}\n"
            f"Response HTTP status is {response.status_code} {response.reason_phrase}\n"
            f"Response payload: {response.text}"
        )
        super().__init__(exception_message)

    @classmethod
    @contextmanager
    def reraise_from(
        cls,
        *exception_types: type[BaseException],
        request: httpx.Request,
        response: httpx.Response,
    ) -> Generator[None, None, None]:
        try:
            yield
        except Exception as exc:  # noqa: BLE001
            if not any(isinstance(exc, exc_type) for exc_type in exception_types):
                raise
            raise cls(request=request, response=response) from exc


def raise_for_status(response: httpx.Response) -> None:
    request = response._request

    if request is None:
        raise ValueError(
            'Cannot call `raise_for_status` as the request instance has not been set on this response.',
        )

    if response.is_success:
        return

    raise HTTPStatusError(request=request, response=response)



from typing import Literal, Union, Any, cast

from pydantic import (
    BaseModel,
    ConstrainedStr,
    ConstrainedInt,
    EmailStr,
    Field,
    conint,
    constr,
    validator,
    root_validator,
    StrictStr,
    StrictFloat,
    StrictInt,
    StrictBool,
)

IntFlag: ConstrainedInt = conint(ge=0, le=1)


AttachmentType: ConstrainedStr = constr(min_length=1, strip_whitespace=True, to_lower=True)
AttachmentName: ConstrainedStr = constr(min_length=1, strip_whitespace=True)


class Attachment(BaseModel):
    type_: AttachmentType = Field(
        alias='type',
        description='Тип вложения, см. MIME https://en.wikipedia.org/wiki/MIME. '
                    'Если не уверены, используйте "application/octet-stream".',
    )
    name: AttachmentName = Field(
        description='Название вложения в формате: "имя.расширение". При передаче нескольких вложений их названия '
                    'должны быть уникальны. В названии вложений запрещен символ `/`.',
    )
    content: str = Field(
        max_length=9786710,
        description='Содержимое файла в base64 https://en.wikipedia.org/wiki/Base64. Максимальный размер '
                    'файла 7Mб (9786710 байт в base64).',
    )

    class Config:
        allow_population_by_field_name = True

    @validator('name')
    @classmethod
    def validate_name(cls, v: str) -> str:
        if '/' in v:
            raise ValueError(f'Symbol `/` is prohibited to use inside attachment name: {v!r}.')

        return v


InlineAttachmentType: ConstrainedStr = constr(min_length=1, strip_whitespace=True, to_lower=True)
InlineAttachmentName: ConstrainedStr = constr(min_length=1, strip_whitespace=True)


class InlineAttachment(BaseModel):
    type_: InlineAttachmentType = Field(
        alias='type',
        description='Тип вложения, см. MIME https://en.wikipedia.org/wiki/MIME. '
                    'Если не уверены, используйте "application/octet-stream".',
    )
    name: InlineAttachmentName = Field(
        description='Идентификатор вложения. Если, например, name="IMAGECID", то для вывода картинки из '
                    'этого вложения в HTML тексте письма надо указать <img src="cid:IMAGECID">.',
    )
    content: str = Field(
        max_length=9786710,
        description='Содержимое файла в base64 https://en.wikipedia.org/wiki/Base64. Максимальный размер '
                    'файла 7Mб (9786710 байт в base64).',
    )

    class Config:
        allow_population_by_field_name = True


SubstitutionName: ConstrainedStr = constr(regex=r'^[a-zA-Z][a-zA-Z0-9\_]*$', strict=True)
SubstitutionValue = Union[StrictStr | StrictFloat | StrictInt | StrictBool | None, 'Substitutions']


class Substitutions(BaseModel):
    __root__: dict[SubstitutionName, SubstitutionValue]


class Recipient(BaseModel):
    email: EmailStr = Field(description='Email-адрес получателя')

    substitutions: Substitutions = Field(
        default_factory=dict,
        description='Объект, описывающий подстановки для конкретного получателя. Например, имя, товар для '
                    'показа именно данному получателю - см. Шаблонизаторы '
                    'https://godocs.unisender.ru/template-engines. Подстановки работают в параметрах: \n'
                    '- body.html\n'
                    '- body.plaintext\n'
                    '- body.amp\n'
                    '- subject\n'
                    '- from_name\n'
                    '- headers["List-Unsubscribe"]\n'
                    '- options.unsubscribe_url\n'
                    'Названия переменных могут содержать буквы латинского алфавита, цифры, символ "_". '
                    'Первым символом может быть только буква, использование пробелов не допускается. '
                    'Есть специальная переменная "to_name", позволяющая задать имя получателя. '
                    'Длина переменной "to_name" ограничена 78 символами. '
                    'При подстановке туда "Имя Фамилия" SMTP-заголовок "To" будет выглядеть как '
                    '"Имя Фамилия <email@example.com>".',
    )

    @validator('substitutions')
    @classmethod
    def validate_to_name_substitution_variable(cls, v: Substitutions) -> Substitutions:
        if 'to_name' not in v.__root__:
            return v

        to_name_value = cast(str, v.__root__['to_name'])

        if len(to_name_value) > 78:
            raise ValueError('Too long to_name value. Should be less than or equal to 78.')

        return v


class EmailBody(BaseModel):
    html: str = ''
    plaintext: str = ''
    amp: str = ''

    @root_validator
    @classmethod
    def validate_content(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not values['html'] and not values['plaintext']:
            raise ValueError(
                'At least one of attributes `html` or `plaintext` should be specified.',
            )

        return values


Tag: ConstrainedStr = constr(max_length=50, min_length=1, strip_whitespace=True)
TemplateId: ConstrainedStr = constr(min_length=1, strip_whitespace=True)


class Message(BaseModel):
    recipients: list[Recipient] = Field(min_items=1)

    template_id: TemplateId | None = Field(
        default=None,
        description='Идентификатор шаблона, созданного ранее методом template/set. Если '
                    'указан, то поля шаблона подставляются вместо пропущенных полей email/send. '
                    'Например, если в email/send не указан body - берётся тело письма из шаблона, '
                    'а если не указан subject - берётся тема из шаблона.',
    )

    tags: list[Tag] = Field(
        default_factory=list,
        max_items=4,
        unique_items=True,
        description='Список уникальных меток, максимум 4 штуки. Метки могут использоваться для '
                    'категоризации писем по выбранным вами критериям. Метки передаются '
                    'при вызове вебхуков и методов event-dump. '
                    'Для каждого аккаунта/проекта разрешено не более 10 000 меток, при превышении '
                    'метод email/send будет возвращать ошибку.',
    )

    skip_unsubscribe: IntFlag = Field(
        default=False,
        description='Пропустить или не пропускать добавление стандартного блока со ссылкой отписки к HTML-части '
                    'письма. Для использования skip_unsubscribe=True вам надо попросить техподдержку включить '
                    'такую возможность.',
    )
    global_language: Literal[
        'be',
        'de',
        'en',
        'es',
        'fr',
        'it',
        'pl',
        'pt',
        'ru',
        'ua',
        'kz',
    ] | None = Field(
        default=None,
        description='Заголовок для выбора языка ссылки и страницы отписки.',
    )

    template_engine: Literal['simple', 'velocity', 'none'] = Field(
        default='simple',
        description='Параметр для выбора шаблонизатора. '
                    'Шаблонизатор `none` доступен только для писем с отключенными `track_links`, '
                    '`track_read` и отключенным блоком отписки.',
    )

    global_substitutions: Substitutions = Field(
        default_factory=dict,
        description='Объект для передачи глобальных подстановок (например, название компании). Если названия '
                    'переменных повторяются в объекте пользовательских подстановок "substitutions", значения '
                    'переменных будут взяты из объекта "substitutions" Подстановки работают в параметрах:\n'
                    '- body.html\n'
                    '- body.plaintext\n'
                    '- body.amp\n'
                    '- subject\n'
                    '- from_name\n'
                    '- options.unsubscribe_url\n'
                    'Названия переменных могут содержать буквы латинского алфавита, цифры, символ "_". '
                    'Первым символом может быть только буква, использование пробелов не допускается.',
    )

    body: EmailBody = Field(
        description='Объект, который содержит в себе html, plaintext и amp части письма. Либо html, либо plaintext '
                    'часть должна присутствовать обязательно.',
    )

    subject: str = Field(
        description='Тема письма.',
    )

    from_email: EmailStr | None = Field(
        default=None,
        description='Email-адрес отправителя. Если значение не указано, то берётся из шаблона.'
                    'Поле обязательно для заполнения только при пустом `template_id`.',
    )
    from_name: str | None = Field(
        default=None,
        description='Имя отправителя. Если значение не указано, то берётся из шаблона.',
    )

    reply_to: EmailStr | None = Field(
        default=None,
        description='Необязательный email-адрес для ответов — на случай, если отличается от адреса отправителя.',
    )
    reply_to_name: str | None = Field(
        default=None,
        description='Необязательное имя для ответов (если указан email reply_to и вы хотите, чтобы отображался '
                    'не только этот email, но и имя).',
    )

    track_links: IntFlag = Field(
        default=True,
        description='Включает/выключает отслеживание переходов по ссылкам.',
    )
    track_read: IntFlag = Field(
        default=True,
        description='Включает/выключает отслеживание прочтений письма.',
    )

    attachments: list[Attachment] = Field(
        default_factory=list,
        description='Вложения к письму',
    )
    inline_attachments: list[InlineAttachment] = Field(
        default_factory=list,
        description='Inline-вложения к письму, например, для включения изображений в письмо вместо '
                    'загрузки их по ссылке.',
    )

    @root_validator
    @classmethod
    def validate_template_engine(cls, values: dict[str, Any]) -> dict[str, Any]:
        none_template_allowed = all([
            not values['track_links'],
            not values['track_read'],
            values['skip_unsubscribe'],
        ])

        if values['template_engine'] == 'none' and not none_template_allowed:
            raise ValueError(
                'Template engine `none` requires options `track_links`, `track_read`'
                'be disabled and option `skip_unsubscribe` be enabled.',
            )

        return values

    @root_validator()
    @classmethod
    def validate_required_tags_if_template_not_specified(cls, values: dict[str, Any]) -> dict[str, Any]:
        if 'template_id' in values:
            return values

        if not values['from_email']:
            raise ValueError('Value `from_email` is required because of empty `template_id`.')

        return values



from collections import Counter
from typing import Literal
from urllib.parse import urljoin

import httpx
from pydantic import BaseModel, Field, validator, ValidationError, Extra

from .exceptions import raise_for_status, SyncClientSetupError, ResponseFormatError
from .clients import SyncClient
from .api_types import Message, Tag


def post_as_json(request_payload: BaseModel, api_method: str) -> httpx.Response:
    """Send a request to the Unisender Go API synchronously using a JSON payload."""
    client = SyncClient.default_client.get(None)

    if not client:
        raise SyncClientSetupError

    http_response = client.session.post(
        urljoin(client.unisender_go_api_root_url, api_method),
        headers={
            'X-API-KEY': client.token,
            'content-type': 'application/json',
            'accept': 'application/json',
        },
        content=request_payload.json(exclude_none=True, by_alias=True).encode('utf-8'),
    )

    raise_for_status(http_response)
    return http_response


class SendRequest(BaseModel):
    """Метод для отправки писем вашим подписчикам.

    При отправке вы можете использовать подстановки переменных, шаблоны, включить отслеживание прочтений или
    переходов по ссылкам и многое другое.

    Оф.документация к методу API: https://godocs.unisender.ru/web-api-ref#email-send .
    """

    message: Message = Field(description='Объект, содержащий все свойства отправляемого сообщения.')

    class Config:
        extra = Extra.forbid
        validate_all = True
        validate_assignment = True

    def send(self) -> 'SendResponse':
        response = post_as_json(self, 'v1/email/send.json')

        with ResponseFormatError.reraise_from(
            ValidationError,
            request=response._request,
            response=response,
        ):
            return SendResponse.parse_raw(response.content)

    @validator('message')
    @classmethod
    def validate_message(cls, v: Message) -> Message:
        emails_counter = Counter(recipient.email for recipient in v.recipients)
        duplicated_emails = [item for item, count in emails_counter.items() if count > 1]
        if duplicated_emails:
            raise ValueError(f'Duplicated emails: {duplicated_emails!r}.')

        emails_number = len(emails_counter)
        if emails_number > 500:
            raise ValueError(f'Emails limit is exceeded {emails_number!r}>500.')

        return v


class SendResponse(BaseModel):
    status: Literal['success']

    job_id: str = Field(
        description='Идентификатор задания по отправке, может оказаться полезным при выяснении причин сбоев.',
    )
    emails: list[str] = Field(
        default_factory=list,
        description='Массив email-адресов, успешно принятых к отправке.',
    )
    tags: list[Tag] = Field(
        default_factory=list,
        description='Незадокументированное в API поле с тегами.',
    )
    failed_emails: dict[str, str] = Field(
        default_factory=dict,
        description='Объект с email-адресами, не принятыми к отправке. '
                    'Email представлен ключом словаря, а статус отправки - значением, '
                    'например: {"email1@gmail.com": "temporary_unavailable"}\n'
                    'Статусы failed_emails описаны в документации к API '
                    'https://godocs.unisender.ru/web-api-ref#email-send',
    )


class ErrorResponse(BaseModel):
    status: Literal['error']
    code: int = Field(
        description='API-код ошибки (не путайте с HTTP-кодом)',
    )
    message: str = Field(
        description='Сообщение об ошибке на английском.',
    )



from .exceptions import (  # noqa: F401
    UnisenderGoError,
    ClientSetupError,
    SyncClientSetupError,
    AsyncClientSetupError,
    ResponseFormatError,
    HTTPStatusError,
    raise_for_status,
)
from .clients import SyncClient  # noqa: F401
from .api_methods import (  # noqa: F401
    post_as_json,
    SendRequest,
    SendResponse,
    ErrorResponse,
)



from contextlib import asynccontextmanager, contextmanager, AsyncExitStack  # noqa F401
from contextvars import ContextVar, Token
from dataclasses import dataclass, KW_ONLY, field
from typing import AsyncGenerator, ClassVar, Generator, Type, TypeVar  # noqa F401

import httpx

DEFAULT_UNISENDER_GO_API_ROOT_URL = 'https://go1.unisender.ru/ru/transactional/api/'

SyncClientType = TypeVar('SyncClientType', bound='SyncClient')


@dataclass(frozen=True)
class SyncClient:
    token: str
    _: KW_ONLY
    session: httpx.Client = field(default_factory=httpx.Client)
    unisender_go_api_root_url: str = DEFAULT_UNISENDER_GO_API_ROOT_URL

    default_client: ClassVar[ContextVar['SyncClient']] = ContextVar('default_client')

    @classmethod
    @contextmanager
    def setup(
        cls: Type[SyncClientType],
        token: str,
        *,
        session: httpx.Client | None = None,
        unisender_go_api_root_url: str = DEFAULT_UNISENDER_GO_API_ROOT_URL,
    ) -> Generator[SyncClientType, None, None]:
        if not token.strip():
            # Safety check for empty string or None to avoid confusing HTTP errors
            raise ValueError(f'Unisender Go token is empty: {token!r}')

        if not session:
            session = httpx.Client()

        client = cls(token=token, session=session, unisender_go_api_root_url=unisender_go_api_root_url)
        with client.set_as_default():
            yield client

    @contextmanager
    def set_as_default(self) -> Generator[None, None, None]:
        default_client_token: Token = self.default_client.set(self)
        try:
            yield
        finally:
            self.default_client.reset(default_client_token)
